class EmptyNormalizationFilter:
    """
    A class to filter out entities with empty normalized values from text data.
    """

    def transform_batch(self, dataset):
        """
        Filters entities with empty normalized values from a given dataset.

        Args:
        - dataset (Dataset): A dataset object from the Hugging Face datasets library.

        Returns:
        - Dataset: A transformed dataset object with the same number of examples, but with empty-normalized entities removed.
        """

        def filter_empty(d):
            ents = [e for e in d["entities"] if len(e["normalized"]) > 0]
            return {"entities": ents}

        return dataset.map(filter_empty)



from .indexed_dataset import IndexedDatasetDict, IndexedDataset
from .util import *
from .integrations import *

from .abbrevations import AbbreviationExpander
from .sampling import Sampler
from .merge_concepts import ConceptMerger
from .retired_cuis import CUIReplacer
from .filter import EmptyNormalizationFilter
from .semantic_types import SemanticTypeFilter
from .semantic_groups import SemanticGroupFilter



def filter_and_apply_threshold(input_pred, k: int, threshold: float):
    """
    Filters the predicted entities based on a score threshold and returns the top k entities for each entity.

    Args:
    - input_pred (list): A list of predicted entities to be filtered and thresholded.
    - k (int): The number of top entities to keep after filtering.
    - threshold (float): The score threshold below which entities will be removed.

    Returns:
    - list: A list of filtered and thresholded predicted entities.
    """
    assert k >= 0

    def apply(entry):
        entities = entry["entities"]
        for e in entities:
            filtered = [n for n in e["normalized"] if "score" in n and n["score"] >= threshold]
            e["normalized"] = sorted(filtered, key=lambda n: n["score"], reverse=True)[:k]
        return {"entities": entities}

    return input_pred.map(apply, load_from_cache_file=False)



import datasets


def from_spacy(docs, span_key=None, doc_id_key=None):
    """
    Converts a list of spaCy documents into a Hugging Face Datasets dataset.

    Args:
    - docs: a list of spaCy documents.
    - span_key: a string specifying the span attribute to use for entity spans. If None, the default 'ents' attribute is used.
    - doc_id_key: a string specifying the key for the document ID to use. If None, document ID is assigned based on position in the input list.

    Returns:
    - a Hugging Face Datasets dataset.
    """
    ds = []
    id_range = list(range(0, len(docs)))

    ds_dict = {
        "id": [],
        "document_id": [],
        "passages": [],
        "entities": [],
    }

    for id, doc in zip(id_range, docs):
        if doc_id_key:
            document_id = doc.user_data[doc_id_key]
        else:
            document_id = id
        ds_dict["id"].append(id)
        ds_dict["document_id"].append(document_id)
        passages = []
        for sid, sent in enumerate(doc.sents):
            passages.append(
                {"id": sid, "type": "sentence", "text": [sent.text], "offsets": [[sent.start_char, sent.end_char]]}
            )
        ds_dict["passages"].append(passages)
        entities = []
        if span_key:
            spans = doc.spans[span_key]
        else:
            spans = doc.ents
        for span_id, span in enumerate(sorted(spans, key=lambda s: s.start_char)):
            entities.append(
                {
                    "id": span_id,
                    "type": span.label_,
                    "text": [span.text],
                    "offsets": [[span.start_char, span.end_char]],
                    "normalized": [],
                }
            )
        ds_dict["entities"].append(entities)
    return datasets.Dataset.from_dict(ds_dict)



from typing import Union, List


def clean_concepts_from_dataset(dataset):
    """
    Cleans the 'entities' key of a given dataset.

    Args:
    - dataset: A dataset containing dictionaries with an 'entities' key.

    Returns:
    - A cleaned dataset where each dictionary in the dataset only contains the 'id', 'spans_start', 'spans_end', and 'text' keys in its 'entities' key.
    """
    return dataset.map(
        lambda i: {
            "entities": {k: v for k, v in i["entities"].items() if k in ["id", "spans_start", "spans_end", "text"]}
        }
    )


class Concept:
    """
    A class representing a concept.

    Args:
    - db_id (str): The id of the concept.
    - score (float): The score of the concept.
    - db_name (str): The name of the knowledge base containing the concept.
    - type (str): The type of the concept.

    Attributes:
    - _dict (dict): A dictionary containing the 'db_id', 'target_kb', 'type', and 'score' keys.
    """

    def __init__(
        self,
        db_id: str = None,
        score: float = None,
        db_name: str = None,
        type: str = None,
    ):
        self._dict = {
            "db_id": db_id,
            "target_kb": db_name,
            "type": type,
            "score": score,
        }


class Entity:
    """
    A class representing an entity.

    Args:
    - offsets: A tuple or list of tuples representing the start and end character offsets of the entity.
    - text (Union[str, List[str]]): A string or list of strings representing the text of the entity.
    - id (str): The id of the entity.
    - entity_type (str): The type of the entity.
    - concepts (List[Concept]): A list of concepts associated with the entity.

    Attributes:
    - _dict (dict): A dictionary containing the 'id', 'text', 'offsets', 'type', and 'normalized' keys.
    """

    def __init__(
        self,
        offsets,
        text: Union[str, List[str]],
        id: str = "1",
        entity_type: str = None,
        concepts: List[Concept] = None,
    ):
        self._dict = {
            "id": id,
            "text": [text] if type(text) == str else text,
            "offsets": offsets,
            "type": entity_type,
            "normalized": [c._dict for c in concepts] if concepts else [],
        }


def make_document(entities: List[Entity], document_id: str = "1", corpus_id: str = "x") -> dict:
    """
    Creates a dictionary representing a document.

    Args:
    - entities (List[Entity]): A list of entities in the document.
    - document_id (str): The id of the document.
    - corpus_id (str): The id of the corpus containing the document.

    Returns:
    - A dictionary with the 'corpus_id', 'document_id', and 'entities' keys.
    """
    return {
        "corpus_id": corpus_id,
        "document_id": document_id,
        "entities": [e._dict for e in entities],
    }


def get_cuis(dataset):
    """
    Extracts the unique CUIs (Concept Unique Identifiers) from a dataset.

    Args:
    - dataset: A dictionary containing the 'entities' key, which is a list of Entity objects.

    Returns:
    - A list of unique CUIs.
    """
    return [c["db_id"] for doc in dataset["entities"] for e in doc for c in e["normalized"]]



import os
import pickle
import logging
import torch
import numpy as np
import time
from tqdm import tqdm
from torch import nn
from sklearn.metrics.pairwise import cosine_similarity
from .metric_learning import *

from transformers import (
    AutoTokenizer, 
    AutoModel, 
)

LOGGER = logging.getLogger()


class Model_Wrapper(object):
    """
    Wrapper class for BERT encoder
    """

    def __init__(self):
        self.tokenizer = None
        self.encoder = None

    def get_dense_encoder(self):
        assert (self.encoder is not None)

        return self.encoder

    def get_dense_tokenizer(self):
        assert (self.tokenizer is not None)

        return self.tokenizer

    def save_model(self, path, context=False):
        # save bert model, bert config
        self.encoder.save_pretrained(path)

        # save bert vocab
        self.tokenizer.save_pretrained(path)
        

    def load_model(self, path, max_length=25, use_cuda=True, lowercase=True):
        self.load_bert(path, max_length, use_cuda)
        
        return self

    def load_bert(self, path, max_length, use_cuda, lowercase=True):
        self.tokenizer = AutoTokenizer.from_pretrained(path, 
                use_fast=True, do_lower_case=lowercase)
        self.encoder = AutoModel.from_pretrained(path)
        if use_cuda:
            self.encoder = self.encoder.cuda()

        return self.encoder, self.tokenizer
    

    def get_score_matrix(self, query_embeds, dict_embeds, cosine=False, normalise=False):
        """
        Return score matrix

        Parameters
        ----------
        query_embeds : np.array
            2d numpy array of query embeddings
        dict_embeds : np.array
            2d numpy array of query embeddings

        Returns
        -------
        score_matrix : np.array
            2d numpy array of scores
        """
        if cosine:
            score_matrix = cosine_similarity(query_embeds, dict_embeds)
        else:
            score_matrix = np.matmul(query_embeds, dict_embeds.T)

        if normalise:
            score_matrix = (score_matrix - score_matrix.min() ) / (score_matrix.max() - score_matrix.min())
        
        return score_matrix

    def retrieve_candidate(self, score_matrix, topk):
        """
        Return sorted topk idxes (descending order)

        Parameters
        ----------
        score_matrix : np.array
            2d numpy array of scores
        topk : int
            The number of candidates

        Returns
        -------
        topk_idxs : np.array
            2d numpy array of scores [# of query , # of dict]
        """
        
        def indexing_2d(arr, cols):
            rows = np.repeat(np.arange(0,cols.shape[0])[:, np.newaxis],cols.shape[1],axis=1)
            return arr[rows, cols]

        # get topk indexes without sorting
        topk_idxs = np.argpartition(score_matrix,-topk)[:, -topk:]

        # get topk indexes with sorting
        topk_score_matrix = indexing_2d(score_matrix, topk_idxs)
        topk_argidxs = np.argsort(-topk_score_matrix) 
        topk_idxs = indexing_2d(topk_idxs, topk_argidxs)

        return topk_idxs

    def retrieve_candidate_cuda(self, score_matrix, topk, batch_size=128, show_progress=False):
        """
        Return sorted topk idxes (descending order)

        Parameters
        ----------
        score_matrix : np.array
            2d numpy array of scores
        topk : int
            The number of candidates

        Returns
        -------
        topk_idxs : np.array
            2d numpy array of scores [# of query , # of dict]
        """

        res = None
        for i in tqdm(np.arange(0, score_matrix.shape[0], batch_size), disable=not show_progress):
            score_matrix_tmp = torch.tensor(score_matrix[i:i+batch_size]).cuda()
            matrix_sorted = torch.argsort(score_matrix_tmp, dim=1, descending=True)[:, :topk].cpu()
            if res is None: 
                res = matrix_sorted
            else:
                res = torch.cat([res, matrix_sorted], axis=0)

        return res.numpy()

    def embed_dense(self, names, show_progress=False, use_cuda=True, batch_size=2048, agg_mode="cls"):
        """
        Embedding data into dense representations

        Parameters
        ----------
        names : np.array
            An array of names

        Returns
        -------
        dense_embeds : list
            A list of dense embeddings
        """
        self.encoder.eval() # prevent dropout
        
        batch_size=batch_size
        dense_embeds = []

        #print ("converting names to list...")
        #names = names.tolist()

        with torch.no_grad():
            if show_progress:
                iterations = tqdm(range(0, len(names), batch_size))
            else:
                iterations = range(0, len(names), batch_size)
                
            for start in iterations:
                end = min(start + batch_size, len(names))
                batch = names[start:end]
                batch_tokenized_names = self.tokenizer.batch_encode_plus(
                        batch, add_special_tokens=True, 
                        truncation=True, max_length=25, 
                        padding="max_length", return_tensors='pt')
                
                # generate cuda names only if use cuda
                # and overwrite original batch_tokenized_names
                if use_cuda:
                    batch_tokenized_names_cuda = {}
                    for k,v in batch_tokenized_names.items(): 
                        batch_tokenized_names_cuda[k] = v.cuda()
                    batch_tokenized_names = batch_tokenized_names_cuda
                
                # batch_tokenized_names is passed (-_cuda has been removed)
                last_hidden_state = self.encoder(**batch_tokenized_names)[0]
                if agg_mode == "cls":
                    batch_dense_embeds = last_hidden_state[:,0,:] # [CLS]
                elif agg_mode == "mean_all_tok":
                    batch_dense_embeds = last_hidden_state.mean(1) # pooling
                elif agg_mode == "mean":
                    batch_dense_embeds = (last_hidden_state * batch_tokenized_names['attention_mask'].unsqueeze(-1)).sum(1) / batch_tokenized_names['attention_mask'].sum(-1).unsqueeze(-1)
                else:
                    print ("no such agg_mode:", agg_mode)

                batch_dense_embeds = batch_dense_embeds.cpu().detach().numpy()
                dense_embeds.append(batch_dense_embeds)
        dense_embeds = np.concatenate(dense_embeds, axis=0)
        
        return dense_embeds
